import uuid
from abc import ABC, abstractmethod
from pathlib import Path

from pytodotxt import Task as PytodoTask
from pytodotxt import TodoTxt

from .models import Task


class TaskRepositoryError(Exception):
    """Base exception for repository errors."""

    pass


class TaskRepository(ABC):
    @abstractmethod
    def get_active_tasks(self) -> list[Task]:
        """Get all active tasks."""
        pass

    @abstractmethod
    def get_completed_tasks(self) -> list[Task]:
        """Get all completed tasks."""
        pass

    @abstractmethod
    def save(self, task: Task) -> None:
        """Save a task (create or update)."""
        pass

    @abstractmethod
    def delete(self, task: Task) -> None:
        """Delete a task."""
        pass


class _TaskWithMeta(Task):
    """Internal wrapper to track persistence details."""

    __slots__ = ("_original_text",)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._original_text: str | None = None


class FileTaskRepository(TaskRepository):
    def __init__(self, todo_file: str, done_file: str):
        self.todo_file = Path(todo_file).resolve()
        self.done_file = Path(done_file).resolve()

        if self.todo_file == self.done_file:
            raise ValueError("todo_file and done_file must be distinct")

        try:
            self._ensure_files_exist()
        except Exception as e:
            raise ValueError(f"Repository files not accessible: {e}") from e

    def _ensure_files_exist(self):
        self._create_file_if_missing(self.todo_file)
        self._create_file_if_missing(self.done_file)

    def _create_file_if_missing(self, path: Path):
        if not path.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
            path.touch()

    def _to_domain(self, pytodo_task: PytodoTask) -> Task:
        # PytodoTask attributes are lists, we need to handle that
        attrs = {}
        pytodo_attrs = getattr(pytodo_task, "attributes", None)
        if pytodo_attrs is not None:
            attrs = pytodo_attrs.copy()

        description = pytodo_task.description or ""

        # Strip attributes from description to prevent duplication
        if attrs:
            for key, values in attrs.items():
                # values is typically a list in pytodotxt
                if isinstance(values, list):
                    for val in values:
                        description = description.replace(f"{key}:{val}", "")
                else:
                    description = description.replace(f"{key}:{values}", "")

            # Clean up extra whitespace
            description = " ".join(description.split())

        task = _TaskWithMeta(
            description=description,
            is_completed=pytodo_task.is_completed or False,
            priority=pytodo_task.priority,
            creation_date=pytodo_task.creation_date,
            completion_date=pytodo_task.completion_date,
            projects=pytodo_task.projects if hasattr(pytodo_task, "projects") else [],
            contexts=pytodo_task.contexts if hasattr(pytodo_task, "contexts") else [],
            attributes=attrs,
        )
        task._original_text = str(pytodo_task)
        return task

    def _to_pytodo(self, task: Task) -> PytodoTask:
        # Construct string representation to create PytodoTask
        t = PytodoTask(task.description)
        t.is_completed = task.is_completed
        t.priority = task.priority
        t.creation_date = task.creation_date
        t.completion_date = task.completion_date

        # Attributes in pytodotxt are tricky, usually parsed from string.
        for k, v in task.attributes.items():
            if isinstance(v, list):
                for item in v:
                    t.add_attribute(k, item)
            else:
                t.add_attribute(k, str(v))

        return t

    def get_active_tasks(self) -> list[Task]:
        """Get all active tasks from todo.txt."""
        try:
            todotxt = TodoTxt(str(self.todo_file))
            todotxt.parse()
            return [self._to_domain(t) for t in todotxt.tasks]
        except Exception as e:
            raise TaskRepositoryError(f"Failed to load active tasks: {e}") from e

    def get_completed_tasks(self) -> list[Task]:
        """Get all completed tasks from done.txt."""
        try:
            donetxt = TodoTxt(str(self.done_file))
            donetxt.parse()
            return [self._to_domain(t) for t in donetxt.tasks]
        except Exception as e:
            raise TaskRepositoryError(f"Failed to load completed tasks: {e}") from e

    def save(self, task: Task) -> None:
        """Save a task (create or update)."""
        try:
            # Generate stable ID if missing
            if "cmid" not in task.attributes:
                task.attributes["cmid"] = uuid.uuid4().hex[:8]

            pytodo_task = self._to_pytodo(task)
            new_text = str(pytodo_task)

            # If we have an ID, try to find and remove by ID first
            removed_by_id = False
            if task.id:
                removed_by_id = self._remove_by_id(self.todo_file, task.id)
                if not removed_by_id:
                    removed_by_id = self._remove_by_id(self.done_file, task.id)

            # If not removed by ID (e.g. legacy task), fallback to original text
            original_text = getattr(task, "_original_text", None)
            if not removed_by_id and original_text:
                # Check if the text actually changed to avoid unnecessary IO
                if original_text == new_text:
                    # However, we might be moving files (completing/uncompleting)
                    # So we still need to check logic below
                    pass

                # Try to remove from both files to be safe (in case it moved)
                self._remove_from_file(self.todo_file, original_text)
                self._remove_from_file(self.done_file, original_text)

            # Determine target file based on current state
            target_file = self.done_file if task.is_completed else self.todo_file

            # Add to target file
            todotxt = TodoTxt(str(target_file))
            todotxt.parse()
            todotxt.tasks.append(pytodo_task)
            todotxt.save()

            # Update original text for future updates
            if isinstance(task, _TaskWithMeta):
                task._original_text = new_text
        except Exception as e:
            raise TaskRepositoryError(f"Failed to save task: {e}") from e
        # If it's a plain Task, we can't attach _original_text unless wrapped.
        # But if we are saving a new task, it might become a legacy update later
        # if we don't track it? But we HAVE generated an ID!
        # So future updates will use ID.

    def delete(self, task: Task) -> None:
        """Delete a task."""
        try:
            # Try to delete by ID first if available
            if task.id:
                removed = self._remove_by_id(self.todo_file, task.id)
                if not removed:
                    removed = self._remove_by_id(self.done_file, task.id)

                if removed:
                    return

            # Fallback to original text
            original_text = getattr(task, "_original_text", None)
            if original_text:
                self._remove_from_file(self.todo_file, original_text)
                self._remove_from_file(self.done_file, original_text)
        except Exception as e:
            raise TaskRepositoryError(f"Failed to delete task: {e}") from e

    def _remove_by_id(self, file_path: Path, task_id: str) -> bool:
        """Helper to remove a task by ID from a file."""
        todotxt = TodoTxt(str(file_path))
        todotxt.parse()

        found = False
        to_remove = []

        for t in todotxt.tasks:
            attrs = getattr(t, "attributes", {})
            if not attrs:
                continue

            val = attrs.get("cmid")
            current_id = None
            if isinstance(val, list) and val:
                current_id = val[0]
            elif isinstance(val, str):
                current_id = val

            if current_id == task_id:
                to_remove.append(t)
                found = True

        if found:
            for t in to_remove:
                todotxt.tasks.remove(t)
            todotxt.save()
            return True
        return False

    def _remove_from_file(self, file_path: Path, task_str: str) -> bool:
        """Helper to remove a task string from a file."""
        todotxt = TodoTxt(str(file_path))
        todotxt.parse()

        found = False
        to_remove = []
        # We match by string representation
        for t in todotxt.tasks:
            if str(t) == task_str:
                to_remove.append(t)
                found = True

        if found:
            for t in to_remove:
                todotxt.tasks.remove(t)
            todotxt.save()
            return True
        return False



"""Sort selection modal screen component."""

from typing import ClassVar

from textual.app import ComposeResult
from textual.containers import Container, Vertical
from textual.screen import Screen
from textual.widgets import Button, Footer, Label


class SortSelectScreen(Screen):
    """Modal screen for selecting a sort attribute."""

    BINDINGS: ClassVar[list] = [
        ("escape", "cancel", "Cancel"),
    ]

    DEFAULT_CSS = """
    SortSelectScreen {
        background: $surface;
    }
    
    SortSelectScreen > Container {
        width: auto;
        height: auto;
        border: solid $accent;
        background: $panel;
        padding: 2 3;
        align: center middle;
    }
    
    SortSelectScreen Label#title {
        width: 100%;
        text-align: center;
        text-style: bold;
        margin-bottom: 2;
    }
    
    SortSelectScreen .button-group {
        width: 1fr;
        height: auto;
    }
    
    SortSelectScreen Button {
        width: 100%;
        margin-bottom: 1;
    }
    
    SortSelectScreen Button:last-of-type {
        margin-bottom: 0;
    }
    """

    def __init__(self, callback=None):
        """Initialize the sort selection screen.

        Args:
            callback: Function to call with selected sort attribute when a
                button is pressed. Callback receives attribute name as string:
                'priority', 'context', 'project', 'due', 'created'
        """
        super().__init__()
        self.callback = callback

    def compose(self) -> ComposeResult:
        """Create the sort selection screen."""
        with Container():
            yield Label("Select sort attribute", id="title")

            with Vertical(classes="button-group"):
                yield Button("Priority (A-Z)", id="sort-priority-btn")
                yield Button("Context (@-tags)", id="sort-context-btn")
                yield Button("Project (+projects)", id="sort-project-btn")
                yield Button("Due Date", id="sort-due-btn")
                yield Button("Created Date", id="sort-created-btn")

        yield Footer()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle button presses."""
        button_id = event.button.id

        # Map button IDs to sort attributes
        button_to_attribute = {
            "sort-priority-btn": "priority",
            "sort-context-btn": "context",
            "sort-project-btn": "project",
            "sort-due-btn": "due",
            "sort-created-btn": "created",
        }

        if button_id in button_to_attribute:
            attribute = button_to_attribute[button_id]
            if self.callback:
                self.callback(attribute)
            self.dismiss()

    def action_cancel(self) -> None:
        """Cancel the screen without applying sort."""
        self.dismiss()



class CheckmateError(Exception):
    """Base exception for all Checkmate errors."""

    pass


class TaskValidationError(CheckmateError):
    """Raised when task data is invalid."""

    pass


class TaskOperationError(CheckmateError):
    """Raised when a task operation fails (persistence, etc.)."""

    pass



"""Confirmation modal screen."""

from typing import ClassVar

from textual.app import ComposeResult
from textual.containers import Container, Horizontal
from textual.screen import ModalScreen
from textual.widgets import Button, Label


class ConfirmScreen(ModalScreen[bool]):
    """A modal screen for confirming actions."""

    BINDINGS: ClassVar[list] = [
        ("escape", "cancel", "Cancel"),
        ("y", "confirm", "Confirm"),
        ("n", "cancel", "Cancel"),
    ]

    DEFAULT_CSS = """
    ConfirmScreen {
        align: center middle;
        background: $surface 50%;
    }

    ConfirmScreen > Container {
        width: auto;
        min-width: 40;
        height: auto;
        padding: 1 2;
        background: $panel;
        border: thick $accent;
    }

    ConfirmScreen Label#message {
        width: 100%;
        content-align: center middle;
        margin: 1 0 2 0;
        text-style: bold;
    }

    ConfirmScreen Horizontal {
        width: 100%;
        height: auto;
        align: center middle;
    }

    ConfirmScreen Button {
        margin: 0 1;
    }
    """

    def __init__(self, message: str = "Are you sure?"):
        super().__init__()
        self.message = message

    def compose(self) -> ComposeResult:
        with Container():
            yield Label(self.message, id="message")
            with Horizontal():
                yield Button("Yes", variant="primary", id="yes-btn")
                yield Button("No", variant="error", id="no-btn")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes-btn":
            self.dismiss(True)
        else:
            self.dismiss(False)

    def action_confirm(self) -> None:
        self.dismiss(True)

    def action_cancel(self) -> None:
        self.dismiss(False)



from datetime import date

import pytest

from checkmate.exceptions import TaskValidationError
from checkmate.repository import FileTaskRepository
from checkmate.services import TodoService


@pytest.fixture
def service(tmp_path):
    todo_file = tmp_path / "todo.txt"
    done_file = tmp_path / "done.txt"
    repo = FileTaskRepository(str(todo_file), str(done_file))
    return TodoService(repo)


def test_create_task(service):
    task = service.create_task("Buy milk")
    assert task.description == "Buy milk"
    assert task.creation_date == date.today()

    tasks = service.get_active_tasks()
    assert len(tasks) == 1
    assert tasks[0].description == "Buy milk"


def test_create_task_with_priority_and_due(service):
    task = service.create_task("Buy milk", priority="A", due_date=date(2025, 12, 31))
    assert task.priority == "A"
    assert task.due_date == date(2025, 12, 31)


def test_complete_task(service):
    task = service.create_task("Buy milk")
    service.complete_task(task)

    assert task.is_completed
    assert task.completion_date == date.today()

    incomplete = service.get_active_tasks()
    assert len(incomplete) == 0

    completed = service.get_completed_tasks()
    assert len(completed) == 1
    assert completed[0].description == "Buy milk"


def test_update_task(service):
    task = service.create_task("Buy milk")
    service.update_task(task, description="Buy cookies", priority="B")

    assert task.description == "Buy cookies"
    assert task.priority == "B"

    tasks = service.get_active_tasks()
    assert tasks[0].description == "Buy cookies"


def test_delete_task(service):
    task = service.create_task("Buy milk")
    service.delete_task(task)

    tasks = service.get_active_tasks()
    assert len(tasks) == 0


def test_create_task_validation(service):
    # Invalid priority
    with pytest.raises(
        TaskValidationError, match="Priority must be a single uppercase letter"
    ):
        service.create_task("Invalid priority", priority="AB")

    with pytest.raises(
        TaskValidationError, match="Priority must be a single uppercase letter"
    ):
        service.create_task("Invalid priority", priority="1")


def test_create_task_empty_description(service):
    with pytest.raises(TaskValidationError, match="Task description cannot be empty"):
        service.create_task("")

    with pytest.raises(TaskValidationError, match="Task description cannot be empty"):
        service.create_task("   ")

    task = service.create_task("Valid")
    with pytest.raises(TaskValidationError, match="Task description cannot be empty"):
        service.update_task(task, description="")



"""Checkmate - A terminal user interface client for todos.txt files."""

import argparse
import sys
from typing import ClassVar

from textual.app import App

from .config import discover_files, load_config_file
from .repository import FileTaskRepository
from .screens import TodoListScreen
from .services import TodoService


class CheckmateApp(App):
    TITLE = "Checkmate"
    CSS_PATH = "checkmate.tcss"

    BINDINGS: ClassVar[list] = [
        ("q", "quit", "Quit"),
    ]

    def __init__(self, service: TodoService):
        super().__init__()
        self.service = service

    async def on_mount(self) -> None:
        """Push the main screen when the app starts."""
        await self.push_screen(TodoListScreen())


def parse_args() -> tuple[str | None, str | None]:
    """Parse command-line arguments.

    Returns:
        Tuple of (todo_file, done_file) paths.
    """
    parser = argparse.ArgumentParser(
        description="Checkmate - A terminal user interface client for todos.txt files."
    )
    parser.add_argument("--todo", help="Path to todo.txt file")
    parser.add_argument("--done", help="Path to done.txt file")

    args = parser.parse_args()
    return args.todo, args.done


def main():
    # Parse command-line arguments
    cli_todo, cli_done = parse_args()

    try:
        # Load configuration from .todo/config
        config = load_config_file()

        # Discover file paths with proper precedence
        todo_file, done_file = discover_files(
            cli_todo_file=cli_todo,
            cli_done_file=cli_done,
            config=config,
        )

        # Launch app with discovered file paths
        repository = FileTaskRepository(todo_file=todo_file, done_file=done_file)
        service = TodoService(repository)
        app = CheckmateApp(service=service)
        app.run()
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



"""Task creation/editing screen component."""

import logging
from datetime import date, datetime, timedelta
from typing import TYPE_CHECKING, ClassVar, cast

if TYPE_CHECKING:
    from ..main import CheckmateApp

from textual.app import ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.events import Key, Resize
from textual.reactive import reactive
from textual.screen import Screen
from textual.widgets import Button, Footer, Input, Label, Static, TextArea

from ..exceptions import TaskOperationError, TaskValidationError
from ..models import Task

logger = logging.getLogger(__name__)


class ResponsiveButtonGroup(Static):
    """Container for buttons that stacks vertically when space is tight."""

    is_vertical = reactive(False, recompose=True)

    DEFAULT_CSS = """
    ResponsiveButtonGroup {
        width: 100%;
        height: auto;
        margin-top: 2;
    }
    
    ResponsiveButtonGroup Button {
        margin-right: 1;
    }
    
    ResponsiveButtonGroup.vertical Button {
        width: 100%;
        margin: 0 0 1 0;
    }
    """

    def compose(self) -> ComposeResult:
        """Create button container with appropriate layout."""
        if self.is_vertical:
            container = Vertical(id="button-container")
        else:
            container = Horizontal(id="button-container")

        with container:
            yield Button("Submit", id="submit-btn", variant="primary")
            yield Button("Cancel", id="cancel-btn", variant="default")

    def watch_is_vertical(self, value: bool) -> None:
        """Update CSS class when layout changes."""
        if value:
            self.add_class("vertical")
        else:
            self.remove_class("vertical")

    def on_mount(self) -> None:
        """Check initial size and update layout."""
        self._check_layout()

    def on_resize(self, event: Resize) -> None:
        """Respond to size changes."""
        self._check_layout()

    def _check_layout(self) -> None:
        """Determine if buttons should stack vertically."""
        # Available width for buttons (accounting for padding)
        available_width = self.size.width - 4  # Account for padding
        # Estimate needed width: 2 buttons at ~10 chars each + spacing
        needed_width = 28

        should_be_vertical = available_width < needed_width

        if self.is_vertical != should_be_vertical:
            self.is_vertical = should_be_vertical


class CreateTaskScreen(Screen):
    """Screen for creating a new task or editing an existing one."""

    BINDINGS: ClassVar[list] = [
        ("escape", "cancel", "Cancel"),
        ("ctrl+s", "submit", "Submit"),
    ]

    DEFAULT_CSS = """
    CreateTaskScreen {
        background: $surface;
    }
    
    CreateTaskScreen > Container {
        width: 100%;
        height: auto;
        border: solid $accent;
        background: $panel;
        padding: 1 2;
    }
    
    CreateTaskScreen Label#title {
        width: 100%;
        text-align: center;
        text-style: bold;
        margin-bottom: 1;
        padding-top: 1;
    }
    
    /* Common form section styling */
    .form-section {
        width: 100%;
        height: auto;
        margin-bottom: 1;
    }

    .form-section .main-label {
        width: 100%;
        text-style: bold;
        padding-bottom: 0;
    }
    
    .form-section Input, .form-section TextArea {
        width: 100%;
    }

    #task-input {
        height: 6;
    }
    
    /* Grid row for side-by-side inputs */
    .form-row {
        width: 100%;
        height: auto;
        margin-bottom: 1;
    }
    
    .priority-col {
        width: 14;
        height: auto;
        margin-right: 2;
    }
    
    .due-date-col {
        width: 20;
        height: auto;
    }

    .help-text {
        color: $text-disabled;
        text-style: italic;
        padding-bottom: 1;
    }
    """

    def __init__(self, task: Task | None = None):
        super().__init__()
        self.editing_task = task

    @property
    def app(self) -> CheckmateApp:
        return cast("CheckmateApp", super().app)

    def compose(self) -> ComposeResult:
        title = "Edit Task" if self.editing_task else "Create New Task"

        with Container():
            yield Label(title, id="title")

            # Description field
            with Vertical(classes="form-section"):
                yield Label("Description", classes="main-label")
                yield Label(
                    "Use [primary]@context[/] and [accent]+project[/] to organize",
                    classes="help-text",
                )
                yield TextArea(id="task-input")

            # Priority and Due date fields in a row
            with Horizontal(classes="form-row"):
                # Priority field
                with Vertical(classes="form-section priority-col"):
                    yield Label("Priority", classes="main-label")
                    yield Label("(A-Z)", classes="help-text")
                    yield Input(id="priority-input", max_length=1, placeholder="A-Z")

                # Due date field
                with Vertical(classes="form-section due-date-col"):
                    yield Label("Due Date", classes="main-label")
                    yield Label("YYYY-MM-DD", classes="help-text")
                    yield Input(id="due-input", placeholder="YYYY-MM-DD")

            # Buttons (responsive)
            yield ResponsiveButtonGroup()

        yield Footer()

    def on_mount(self) -> None:
        """Focus description input on mount and prepopulate if editing."""
        task_input = self.query_one("#task-input", TextArea)
        priority_input = self.query_one("#priority-input", Input)
        due_input = self.query_one("#due-input", Input)

        # Prepopulate fields if editing
        if self.editing_task:
            task_input.text = self.editing_task.description
            if self.editing_task.priority:
                priority_input.value = self.editing_task.priority
            # Extract due date
            if self.editing_task.due_date:
                due_input.value = self.editing_task.due_date.strftime("%Y-%m-%d")

        task_input.focus()

    def on_key(self, event: Key) -> None:
        """Handle key presses for the due date input."""
        if not self.focused or self.focused.id != "due-input":
            return

        if event.key in ("up", "down", "shift+up", "shift+down"):
            event.stop()
            self._adjust_due_date(event.key)

    def _adjust_due_date(self, key: str) -> None:
        """Adjust the due date based on key press."""
        due_input = self.query_one("#due-input", Input)
        current_value = due_input.value.strip()

        target_date = date.today()

        if current_value:
            try:
                target_date = datetime.strptime(current_value, "%Y-%m-%d").date()
            except ValueError:
                # If invalid date, stick with today as base
                pass

        delta_days = 0
        if key == "up":
            delta_days = 1
        elif key == "down":
            delta_days = -1
        elif key == "shift+up":
            delta_days = 7
        elif key == "shift+down":
            delta_days = -7

        new_date = target_date + timedelta(days=delta_days)
        due_input.value = new_date.strftime("%Y-%m-%d")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle button presses."""
        if event.button.id == "submit-btn":
            self.action_submit()
        elif event.button.id == "cancel-btn":
            self.action_cancel()

    def on_input_submitted(self, event: Input.Submitted) -> None:
        """Handle Enter key in input field."""
        # TextArea handles its own Enter key (for newlines)
        # so we only handle submission for single-line inputs here
        pass

    def action_submit(self) -> None:
        """Submit the task."""

        priority_input = self.query_one("#priority-input", Input)
        task_input = self.query_one("#task-input", TextArea)
        due_input = self.query_one("#due-input", Input)

        priority = priority_input.value.strip() or None
        # Flatten newlines to spaces for todo.txt compatibility
        task_text = task_input.text.strip().replace("\n", " ")
        due_date = due_input.value.strip() or None

        parsed_due_date = None
        if due_date:
            try:
                parsed_due_date = datetime.strptime(due_date, "%Y-%m-%d").date()
            except ValueError:
                self.notify("Date must be in YYYY-MM-DD format", severity="error")
                due_input.focus()
                return

        if not task_text:
            self.notify("Task description cannot be empty", severity="error")
            task_input.focus()
            return

        try:
            if self.editing_task:
                # Update task
                self.app.service.update_task(
                    self.editing_task,
                    description=task_text,
                    priority=priority,
                    due_date=parsed_due_date,
                )
                # We return the updated task description for notification
                self.dismiss(
                    result={
                        "success": True,
                        "task": self.editing_task.description,
                        "mode": "edit",
                    }
                )
            else:
                # Create new task
                new_task = self.app.service.create_task(
                    description=task_text, priority=priority, due_date=parsed_due_date
                )
                self.dismiss(
                    result={
                        "success": True,
                        "task": new_task.description,
                        "mode": "create",
                    }
                )
        except TaskValidationError as e:
            self.notify(str(e), severity="error")
            if "Priority" in str(e):
                priority_input.focus()
            elif "description" in str(e).lower():
                task_input.focus()
        except TaskOperationError as e:
            self.notify(f"Operation failed: {e}", severity="error")
        except Exception as e:
            logger.exception("Unexpected error in CreateTaskScreen")
            self.notify(f"Unexpected error: {e}", severity="error")

    def action_cancel(self) -> None:
        """Cancel the screen."""
        self.dismiss(result={"success": False, "cancelled": True})



import pytest

from checkmate.models import Task
from checkmate.repository import FileTaskRepository, TaskRepositoryError


@pytest.fixture
def repo(tmp_path):
    todo_file = tmp_path / "todo.txt"
    done_file = tmp_path / "done.txt"
    return FileTaskRepository(str(todo_file), str(done_file))


def test_repository_raises_error_on_io_failure(repo, tmp_path):
    # Make file read-only to trigger IO error
    p = tmp_path / "todo.txt"
    p.touch()
    p.chmod(0o000)

    try:
        with pytest.raises(TaskRepositoryError):
            repo.get_active_tasks()
    finally:
        p.chmod(0o666)  # restore


def test_repository_validates_distinct_paths(tmp_path):
    f = tmp_path / "tasks.txt"
    with pytest.raises(ValueError, match="distinct"):
        FileTaskRepository(str(f), str(f))


def test_repository_validates_accessibility(tmp_path):
    # Create a directory that is read-only
    d = tmp_path / "readonly"
    d.mkdir()
    try:
        # Remove write permission
        d.chmod(0o444)

        # Attempt to create files inside readonly dir
        with pytest.raises(ValueError, match="accessible"):
            FileTaskRepository(str(d / "todo.txt"), str(d / "done.txt"))
    finally:
        # Restore permission to allow cleanup
        d.chmod(0o777)


def test_save_generates_id(repo, tmp_path):
    task = Task("Test task")
    repo.save(task)

    # Verify task has an ID in attributes
    assert "cmid" in task.attributes

    # Verify ID is persisted to file
    content = (tmp_path / "todo.txt").read_text()
    assert f"cmid:{task.attributes['cmid']}" in content


def test_id_is_stable(repo):
    task = Task("Test task")
    repo.save(task)
    initial_id = task.attributes["cmid"]

    task.description = "Updated task"
    repo.save(task)

    assert task.attributes["cmid"] == initial_id



"""Main todo list screen."""

import logging
from typing import TYPE_CHECKING, ClassVar, cast

if TYPE_CHECKING:
    from ..main import CheckmateApp

from textual.app import ComposeResult
from textual.command import DiscoveryHit, Hit, Hits, Provider
from textual.containers import Vertical
from textual.screen import Screen
from textual.widgets import Footer, Header

from ..exceptions import TaskOperationError
from ..widgets.task_list import CompletedTaskList, TaskList
from .confirm import ConfirmScreen
from .create_task import CreateTaskScreen

logger = logging.getLogger(__name__)


class SortCommandProvider(Provider):
    """Provider for task sorting commands."""

    async def discover(self) -> Hits:
        """Discover commands to show by default."""
        screen = cast("TodoListScreen", self.screen)
        task_list = screen.task_list
        if task_list:
            yield DiscoveryHit(
                "Sort by Priority",
                lambda: task_list.apply_sort("priority"),
                help="Sort tasks by priority",
            )
            yield DiscoveryHit(
                "Sort by Context",
                lambda: task_list.apply_sort("context"),
                help="Sort tasks by context",
            )
            yield DiscoveryHit(
                "Sort by Project",
                lambda: task_list.apply_sort("project"),
                help="Sort tasks by project",
            )
            yield DiscoveryHit(
                "Sort by Due Date",
                lambda: task_list.apply_sort("due"),
                help="Sort tasks by due date",
            )
            yield DiscoveryHit(
                "Sort by Created Date",
                lambda: task_list.apply_sort("created"),
                help="Sort tasks by creation date",
            )

    async def search(self, query: str) -> Hits:
        """Search for commands."""
        matcher = self.matcher(query)
        screen = cast("TodoListScreen", self.screen)
        task_list = screen.task_list

        if not task_list:
            return

        commands = [
            (
                "Sort by Priority",
                lambda: task_list.apply_sort("priority"),
                "Sort tasks by priority",
            ),
            (
                "Sort by Context",
                lambda: task_list.apply_sort("context"),
                "Sort tasks by context",
            ),
            (
                "Sort by Project",
                lambda: task_list.apply_sort("project"),
                "Sort tasks by project",
            ),
            (
                "Sort by Due Date",
                lambda: task_list.apply_sort("due"),
                "Sort tasks by due date",
            ),
            (
                "Sort by Created Date",
                lambda: task_list.apply_sort("created"),
                "Sort tasks by creation date",
            ),
        ]

        for name, callback, help_text in commands:
            score = matcher.match(name)
            if score > 0:
                yield Hit(
                    score,
                    matcher.highlight(name),
                    callback,
                    help=help_text,
                )


class TodoListScreen(Screen):
    """Primary screen showing the todo and completed task lists."""

    COMMANDS: ClassVar[set] = {SortCommandProvider}

    BINDINGS: ClassVar[list] = [
        ("a", "add_todo", "Add Todo"),
        ("d", "delete_todo", "Delete Todo"),
        ("D", "force_delete_todo", "Force Delete"),
        ("e", "edit_todo", "Edit Todo"),
        ("C", "toggle_completed", "Toggle Completed"),
        ("1", "focus_todo", "Todo List"),
        ("2", "focus_completed", "Done List"),
    ]

    def __init__(self):
        super().__init__()
        self.task_list: TaskList | None = None
        self.completed_list: CompletedTaskList | None = None

    @property
    def app(self) -> CheckmateApp:
        return cast("CheckmateApp", super().app)

    def compose(self) -> ComposeResult:
        yield Header()
        self.task_list = TaskList(id="task-list")
        self.completed_list = CompletedTaskList(id="completed-list")
        with Vertical(id="todos-container"):
            yield self.task_list
            yield self.completed_list
        yield Footer()

    def action_add_todo(self) -> None:
        """Show the task creation modal."""

        def on_modal_result(result):
            if result and result.get("success"):
                self.app.notify(f"Task created: {result.get('task')}", timeout=2.0)
                if self.task_list:
                    self.task_list.refresh_tasks()
            elif result and result.get("error"):
                self.app.notify(
                    f"Error: {result.get('error')}", severity="error", timeout=5.0
                )

        self.app.push_screen(CreateTaskScreen(), callback=on_modal_result)

    def _delete_task(self) -> None:
        """Helper to delete the task at cursor."""
        if self.task_list and self.task_list.has_focus:
            task = self.task_list.get_task_at_cursor()
            if task:
                try:
                    self.task_list.delete_task_at_cursor()
                    self.app.notify(f"Task deleted: {task.description}", timeout=2.0)
                except TaskOperationError as e:
                    self.app.notify(
                        f"Failed to delete task: {e}", severity="error", timeout=5.0
                    )
                except Exception as e:
                    logger.exception("Unexpected error deleting task")
                    self.app.notify(
                        f"Unexpected error: {e}", severity="error", timeout=5.0
                    )
            else:
                self.app.notify("No task selected", severity="warning", timeout=2.0)
        elif self.completed_list and self.completed_list.has_focus:
            self.app.notify(
                "Deleting completed tasks is not supported",
                severity="warning",
                timeout=2.0,
            )

    def action_delete_todo(self) -> None:
        """Delete the currently selected task with confirmation."""
        if self.task_list and self.task_list.has_focus:
            task = self.task_list.get_task_at_cursor()
            if task:

                def on_confirm(result: bool | None) -> None:
                    if result:
                        self._delete_task()

                self.app.push_screen(
                    ConfirmScreen(
                        message=(
                            "Are you sure you want to delete task:\n"
                            f"'{task.description}'?"
                        )
                    ),
                    callback=on_confirm,
                )
            else:
                self.app.notify("No task selected", severity="warning", timeout=2.0)
        elif self.completed_list and self.completed_list.has_focus:
            self.app.notify(
                "Deleting completed tasks is not supported",
                severity="warning",
                timeout=2.0,
            )

    def action_force_delete_todo(self) -> None:
        """Delete the currently selected task without confirmation."""
        self._delete_task()

    def action_edit_todo(self) -> None:
        """Edit the currently selected task."""
        if self.task_list and self.task_list.has_focus:
            task = self.task_list.get_task_at_cursor()
            if task:

                def on_modal_result(result):
                    if result and result.get("success"):
                        self.app.notify(
                            f"Task updated: {result.get('task')}", timeout=2.0
                        )
                        if self.task_list:
                            self.task_list.refresh_tasks()
                    elif result and result.get("error"):
                        self.app.notify(
                            f"Error: {result.get('error')}",
                            severity="error",
                            timeout=5.0,
                        )

                self.app.push_screen(
                    CreateTaskScreen(task=task),
                    callback=on_modal_result,
                )
            else:
                self.app.notify("No task selected", severity="warning", timeout=2.0)
        elif self.completed_list and self.completed_list.has_focus:
            self.app.notify(
                "Editing completed tasks is not supported",
                severity="warning",
                timeout=2.0,
            )

    def action_focus_todo(self) -> None:
        """Focus the todo list."""
        if self.task_list:
            self.task_list.focus()

    def action_focus_completed(self) -> None:
        """Focus the completed list if visible."""
        if self.completed_list and self.completed_list.has_class("visible"):
            self.completed_list.focus()
        elif self.completed_list:
            self.app.notify("Completed list is hidden", severity="warning", timeout=2.0)

    def action_toggle_completed(self) -> None:
        """Toggle visibility of completed tasks list."""
        if self.completed_list:
            if self.completed_list.has_class("visible"):
                self.completed_list.remove_class("visible")
            else:
                self.completed_list.refresh_tasks()
                self.completed_list.add_class("visible")



"""Tests for the sort selection modal screen."""

import pytest
from textual.app import App, ComposeResult
from textual.widgets import Button

from checkmate.screens.sort_select import SortSelectScreen


@pytest.mark.asyncio
async def test_sort_select_screen_compose():
    """Test that SortSelectScreen composes with all button options."""

    class TestApp(App):
        def compose(self) -> ComposeResult:
            yield SortSelectScreen(callback=lambda x: None)

    app = TestApp()
    async with app.run_test() as _:
        app.query_one(SortSelectScreen)

        # Query buttons by their IDs
        buttons = app.query("Button")
        button_ids = [b.id for b in buttons]

        assert "sort-priority-btn" in button_ids
        assert "sort-context-btn" in button_ids
        assert "sort-project-btn" in button_ids
        assert "sort-due-btn" in button_ids
        assert "sort-created-btn" in button_ids


@pytest.mark.asyncio
async def test_sort_select_priority_button():
    """Test clicking the Priority sort button."""
    callback_results = []

    def test_callback(attr):
        callback_results.append(attr)

    class TestApp(App):
        def compose(self) -> ComposeResult:
            yield SortSelectScreen(callback=test_callback)

    app = TestApp()
    async with app.run_test() as pilot:
        button = app.query_one("#sort-priority-btn", Button)
        button.press()
        await pilot.pause()

        assert "priority" in callback_results


@pytest.mark.asyncio
async def test_sort_select_context_button():
    """Test clicking the Context sort button."""
    callback_results = []

    def test_callback(attr):
        callback_results.append(attr)

    class TestApp(App):
        def compose(self) -> ComposeResult:
            yield SortSelectScreen(callback=test_callback)

    app = TestApp()
    async with app.run_test() as pilot:
        button = app.query_one("#sort-context-btn", Button)
        button.press()
        await pilot.pause()

        assert "context" in callback_results


@pytest.mark.asyncio
async def test_sort_select_project_button():
    """Test clicking the Project sort button."""
    callback_results = []

    def test_callback(attr):
        callback_results.append(attr)

    class TestApp(App):
        def compose(self) -> ComposeResult:
            yield SortSelectScreen(callback=test_callback)

    app = TestApp()
    async with app.run_test() as pilot:
        button = app.query_one("#sort-project-btn", Button)
        button.press()
        await pilot.pause()

        assert "project" in callback_results


@pytest.mark.asyncio
async def test_sort_select_due_button():
    """Test clicking the Due date sort button."""
    callback_results = []

    def test_callback(attr):
        callback_results.append(attr)

    class TestApp(App):
        def compose(self) -> ComposeResult:
            yield SortSelectScreen(callback=test_callback)

    app = TestApp()
    async with app.run_test() as pilot:
        button = app.query_one("#sort-due-btn", Button)
        button.press()
        await pilot.pause()

        assert "due" in callback_results


@pytest.mark.asyncio
async def test_sort_select_created_button():
    """Test clicking the Created date sort button."""
    callback_results = []

    def test_callback(attr):
        callback_results.append(attr)

    class TestApp(App):
        def compose(self) -> ComposeResult:
            yield SortSelectScreen(callback=test_callback)

    app = TestApp()
    async with app.run_test() as pilot:
        button = app.query_one("#sort-created-btn", Button)
        button.press()
        await pilot.pause()

        assert "created" in callback_results


@pytest.mark.asyncio
async def test_sort_select_escape_closes():
    """Test that Escape key closes the screen without callback."""
    callback_called = False

    def test_callback(attr):
        nonlocal callback_called
        callback_called = True

    class TestApp(App):
        def compose(self) -> ComposeResult:
            yield SortSelectScreen(callback=test_callback)

    app = TestApp()
    async with app.run_test() as pilot:
        app.query_one(SortSelectScreen)

        # Press escape
        await pilot.press("escape")
        await pilot.pause()

        # Callback should not have been called
        assert not callback_called
